#!/usr/bin/env python

"""Tests for `octis` package."""

import pytest

from octis.dataset.dataset import Dataset
from octis.models.LDA import LDA
from octis.models.LDA_tomopy import LDA_tomopy as LDATOMOTO
from octis.models.ETM import ETM
from octis.models.CTM import CTM
from octis.models.NMF import NMF
from octis.models.NMF_scikit import NMF_scikit
from octis.models.ProdLDA import ProdLDA
from octis.preprocessing.preprocessing import Preprocessing

import numpy as np
import os


@pytest.fixture
def root_dir():
    return os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def data_dir(root_dir):
    return root_dir + "/../preprocessed_datasets/"


@pytest.fixture
def embeddings_dir(root_dir):
    return root_dir + "/../trained_embeddings/"


def test_model_output_lda(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = LDA(num_topics=num_topics, iterations=5)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(
        dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(
        dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (num_topics, len(
        dataset.get_partitioned_corpus()[2]))


def test_model_output_etm(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ETM(num_topics=num_topics, num_epochs=5)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(
        dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (num_topics, len(
        dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (num_topics, len(
        dataset.get_partitioned_corpus()[2]))


def test_model_output_etm_with_corpus_containing_single_word_document(
        data_dir):
    texts_path = data_dir+"/sample_texts/unprepr_docs.txt"
    p = Preprocessing(
        vocabulary=None, max_features=None, remove_punctuation=True,
        lemmatize=False, stopword_list='english')
    dataset = p.preprocess_dataset(
        documents_path=texts_path,
    )
    num_topics = 3
    model = ETM(num_topics=num_topics, num_epochs=5)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(
        dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_etm_not_partitioned(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ETM(num_topics=num_topics, num_epochs=5, use_partitions=False)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' not in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_corpus()))


def test_model_output_etm_with_pickle_embeddings_file(
        data_dir, embeddings_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ETM(
        num_topics=num_topics, num_epochs=5, train_embeddings=False,
        embeddings_path=embeddings_dir + '/test_example/example.pickle')
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_etm_with_binary_word2vec_embeddings_file(
        data_dir, embeddings_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ETM(
        num_topics=num_topics, num_epochs=5, train_embeddings=False,
        embeddings_type='word2vec',
        embeddings_path=embeddings_dir + '/test_example/example.bin')
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[2]))

def test_model_output_etm_with_text_word2vec_embeddings_file(
        data_dir, embeddings_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ETM(
        num_topics=num_topics, num_epochs=5, train_embeddings=False,
        embeddings_type='word2vec',
        embeddings_path=embeddings_dir +'/test_example/example.txt',
        binary_embeddings=False)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_etm_with_headerless_text_word2vec_embeddings_file(
        data_dir, embeddings_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ETM(
        num_topics=num_topics, num_epochs=5, train_embeddings=False,
        embeddings_type='word2vec',
        embeddings_path=embeddings_dir + '/test_example/headerless_example.txt',
        binary_embeddings=False, headerless_embeddings=True)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_etm_with_keyedvectors_embeddings_file(
        data_dir, embeddings_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ETM(
        num_topics=num_topics, num_epochs=5, train_embeddings=False,
        embeddings_type='keyedvectors',
        embeddings_path=embeddings_dir +'/test_example/example.keyedvectors')
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_lda_tomotopy(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = LDATOMOTO(num_topics=num_topics, alpha=0.1)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_nmf(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = NMF(
        num_topics=num_topics, w_max_iter=10, h_max_iter=10,
        use_partitions=True)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_nmf_scikit(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = NMF_scikit(num_topics=num_topics, use_partitions=True)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[2]))


@pytest.mark.skip(reason='the test fails on github actions')
def test_model_output_ctm_zeroshot_seeded(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = CTM(
        num_topics=num_topics, seed=10,
        num_epochs=5, inference_type='zeroshot')
    output = model.train_model(dataset)

    model_2 = CTM(
        num_topics=num_topics, seed=10,
        num_epochs=5, inference_type='zeroshot')
    output_2 = model_2.train_model(dataset)

    assert output['topics'] == output_2['topics']


def test_model_output_ctm_zeroshot(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = CTM(num_topics=num_topics, num_epochs=5, inference_type='zeroshot')
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_ctm_combined(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = CTM(num_topics=num_topics, num_epochs=5, inference_type='combined')
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (num_topics, len(
        dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_ctm_combined_not_partition(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = CTM(
        num_topics=num_topics, num_epochs=5, inference_type='combined',
        use_partitions=False, bert_path='./not_part')
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' not in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_corpus()))


def test_model_output_prodlda(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ProdLDA(num_topics=num_topics, num_epochs=5)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[0]))

    # check test-topic-document-matrix format
    assert type(output['test-topic-document-matrix']) == np.ndarray
    assert output['test-topic-document-matrix'].shape == (
        num_topics, len(dataset.get_partitioned_corpus()[2]))


def test_model_output_prodlda_not_partitioned(data_dir):
    dataset = Dataset()
    dataset.load_custom_dataset_from_folder(data_dir + '/M10')
    num_topics = 3
    model = ProdLDA(num_topics=num_topics, num_epochs=5, use_partitions=False)
    output = model.train_model(dataset)
    assert 'topics' in output.keys()
    assert 'topic-word-matrix' in output.keys()
    assert 'test-topic-document-matrix' not in output.keys()

    # check topics format
    assert type(output['topics']) == list
    assert len(output['topics']) == num_topics

    # check topic-word-matrix format
    assert type(output['topic-word-matrix']) == np.ndarray
    assert output['topic-word-matrix'].shape == (
        num_topics, len(dataset.get_vocabulary()))

    # check topic-document-matrix format
    assert type(output['topic-document-matrix']) == np.ndarray
    assert output['topic-document-matrix'].shape == (
        num_topics, len(dataset.get_corpus()))



"""This file defines a dynamic etm object.
"""
import torch
import torch.nn.functional as F
import numpy as np
import math
from torch import nn


class DETM(nn.Module):
    def __init__(self, num_topics, num_times, vocab_size, t_hidden_size, eta_hidden_size, eta_dropout,
                 rho_size, emb_size, enc_drop, eta_nlayers, delta, train_embeddings,
                 theta_act, embeddings):
        super(DETM, self).__init__()

        ## define hyperparameters
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.num_topics = num_topics
        self.num_times = num_times
        self.vocab_size = vocab_size
        self.t_hidden_size = t_hidden_size
        self.eta_hidden_size = eta_hidden_size
        self.rho_size = rho_size
        self.emsize = emb_size
        self.enc_drop = enc_drop
        self.eta_nlayers = eta_nlayers
        self.t_drop = nn.Dropout(enc_drop)
        self.delta = delta
        self.train_embeddings = train_embeddings

        self.theta_act = self.get_activation(theta_act)

        ## define the word embedding matrix \rho
        if train_embeddings:
            self.rho = nn.Linear(rho_size, vocab_size, bias=False)
        else:
            num_embeddings, emsize = embeddings.size()
            rho = nn.Embedding(num_embeddings, emsize)
            rho.weight.data = embeddings
            self.rho = rho.weight.data.clone().float().to(self.device)

        ## define the variational parameters for the topic embeddings over time (alpha) ... alpha is K x T x L
        self.mu_q_alpha = nn.Parameter(torch.randn(num_topics, num_times, rho_size))
        self.logsigma_q_alpha = nn.Parameter(torch.randn(num_topics, num_times, rho_size))

        ## define variational distribution for \theta_{1:D} via amortizartion... theta is K x D
        self.q_theta = nn.Sequential(
                    nn.Linear(vocab_size+num_topics, t_hidden_size),
                    self.theta_act,
                    nn.Linear(t_hidden_size, t_hidden_size),
                    theta_act,
                )
        self.mu_q_theta = nn.Linear(t_hidden_size, num_topics, bias=True)
        self.logsigma_q_theta = nn.Linear(t_hidden_size, num_topics, bias=True)

        ## define variational distribution for \eta via amortizartion... eta is K x T
        self.q_eta_map = nn.Linear(vocab_size, eta_hidden_size)
        self.q_eta = nn.LSTM(eta_hidden_size, eta_hidden_size, eta_nlayers, dropout=eta_dropout)
        self.mu_q_eta = nn.Linear(eta_hidden_size+num_topics, num_topics, bias=True)
        self.logsigma_q_eta = nn.Linear(eta_hidden_size+num_topics, num_topics, bias=True)

    def get_activation(self, act):
        if act == 'tanh':
            act = nn.Tanh()
        elif act == 'relu':
            act = nn.ReLU()
        elif act == 'softplus':
            act = nn.Softplus()
        elif act == 'rrelu':
            act = nn.RReLU()
        elif act == 'leakyrelu':
            act = nn.LeakyReLU()
        elif act == 'elu':
            act = nn.ELU()
        elif act == 'selu':
            act = nn.SELU()
        elif act == 'glu':
            act = nn.GLU()
        else:
            print('Defaulting to tanh activations...')
            act = nn.Tanh()
        return act

    def reparameterize(self, mu, logvar):
        """Returns a sample from a Gaussian distribution via reparameterization.
        """
        if self.training:
            std = torch.exp(0.5 * logvar)
            eps = torch.randn_like(std)
            return eps.mul_(std).add_(mu)
        else:
            return mu

    def get_kl(self, q_mu, q_logsigma, p_mu=None, p_logsigma=None):
        """Returns KL( N(q_mu, q_logsigma) || N(p_mu, p_logsigma) ).
        """
        if p_mu is not None and p_logsigma is not None:
            sigma_q_sq = torch.exp(q_logsigma)
            sigma_p_sq = torch.exp(p_logsigma)
            kl = ( sigma_q_sq + (q_mu - p_mu)**2 ) / ( sigma_p_sq + 1e-6 )
            kl = kl - 1 + p_logsigma - q_logsigma
            kl = 0.5 * torch.sum(kl, dim=-1)
        else:
            kl = -0.5 * torch.sum(1 + q_logsigma - q_mu.pow(2) - q_logsigma.exp(), dim=-1)
        return kl

    def get_alpha(self): ## mean field
        alphas = torch.zeros(self.num_times, self.num_topics, self.rho_size).to(self.device)
        kl_alpha = []

        alphas[0] = self.reparameterize(self.mu_q_alpha[:, 0, :], self.logsigma_q_alpha[:, 0, :])

        p_mu_0 = torch.zeros(self.num_topics, self.rho_size).to(self.device)
        logsigma_p_0 = torch.zeros(self.num_topics, self.rho_size).to(self.device)
        kl_0 = self.get_kl(self.mu_q_alpha[:, 0, :], self.logsigma_q_alpha[:, 0, :], p_mu_0, logsigma_p_0)
        kl_alpha.append(kl_0)
        for t in range(1, self.num_times):
            alphas[t] = self.reparameterize(self.mu_q_alpha[:, t, :], self.logsigma_q_alpha[:, t, :])

            p_mu_t = alphas[t-1]
            logsigma_p_t = torch.log(self.delta * torch.ones(self.num_topics, self.rho_size).to(self.device))
            kl_t = self.get_kl(self.mu_q_alpha[:, t, :], self.logsigma_q_alpha[:, t, :], p_mu_t, logsigma_p_t)
            kl_alpha.append(kl_t)
        kl_alpha = torch.stack(kl_alpha).sum()
        return alphas, kl_alpha.sum()

    def get_eta(self, rnn_inp): ## structured amortized inference
        inp = self.q_eta_map(rnn_inp).unsqueeze(1)
        hidden = self.init_hidden()
        output, _ = self.q_eta(inp, hidden)
        output = output.squeeze()

        etas = torch.zeros(self.num_times, self.num_topics).to(self.device)
        kl_eta = []

        inp_0 = torch.cat([output[0], torch.zeros(self.num_topics,).to(self.device)], dim=0)
        mu_0 = self.mu_q_eta(inp_0)
        logsigma_0 = self.logsigma_q_eta(inp_0)
        etas[0] = self.reparameterize(mu_0, logsigma_0)

        p_mu_0 = torch.zeros(self.num_topics,).to(self.device)
        logsigma_p_0 = torch.zeros(self.num_topics,).to(self.device)
        kl_0 = self.get_kl(mu_0, logsigma_0, p_mu_0, logsigma_p_0)
        kl_eta.append(kl_0)
        for t in range(1, self.num_times):
            inp_t = torch.cat([output[t], etas[t-1]], dim=0)
            mu_t = self.mu_q_eta(inp_t)
            logsigma_t = self.logsigma_q_eta(inp_t)
            etas[t] = self.reparameterize(mu_t, logsigma_t)

            p_mu_t = etas[t-1]
            logsigma_p_t = torch.log(self.delta * torch.ones(self.num_topics,).to(self.device))
            kl_t = self.get_kl(mu_t, logsigma_t, p_mu_t, logsigma_p_t)
            kl_eta.append(kl_t)
        kl_eta = torch.stack(kl_eta).sum()
        return etas, kl_eta

    def get_theta(self, eta, bows, times): ## amortized inference
        """Returns the topic proportions.
        """
        eta_td = eta[times.type('torch.LongTensor')]
        inp = torch.cat([bows, eta_td], dim=1)
        q_theta = self.q_theta(inp)
        if self.enc_drop > 0:
            q_theta = self.t_drop(q_theta)
        mu_theta = self.mu_q_theta(q_theta)
        logsigma_theta = self.logsigma_q_theta(q_theta)
        z = self.reparameterize(mu_theta, logsigma_theta)
        theta = F.softmax(z, dim=-1)
        kl_theta = self.get_kl(mu_theta, logsigma_theta, eta_td, torch.zeros(self.num_topics).to(self.device))
        return theta, kl_theta

    def get_beta(self, alpha):
        """Returns the topic matrix \beta of shape K x V
        """
        if self.train_embeddings:
            logit = self.rho(alpha.view(alpha.size(0)*alpha.size(1), self.rho_size))
        else:
            tmp = alpha.view(alpha.size(0)*alpha.size(1), self.rho_size)
            logit = torch.mm(tmp, self.rho.permute(1, 0))
        logit = logit.view(alpha.size(0), alpha.size(1), -1)
        beta = F.softmax(logit, dim=-1)
        return beta

    def get_nll(self, theta, beta, bows):
        theta = theta.unsqueeze(1)
        loglik = torch.bmm(theta, beta).squeeze(1)
        loglik = loglik
        loglik = torch.log(loglik+1e-6)
        nll = -loglik * bows
        nll = nll.sum(-1)
        return nll

    def forward(self, bows, normalized_bows, times, rnn_inp, num_docs):
        bsz = normalized_bows.size(0)
        coeff = num_docs / bsz
        alpha, kl_alpha = self.get_alpha()
        eta, kl_eta = self.get_eta(rnn_inp)
        theta, kl_theta = self.get_theta(eta, normalized_bows, times)
        kl_theta = kl_theta.sum() * coeff

        beta = self.get_beta(alpha)
        beta = beta[times.type('torch.LongTensor')]
        nll = self.get_nll(theta, beta, bows)
        nll = nll.sum() * coeff
        nelbo = nll + kl_alpha + kl_eta + kl_theta
        return nelbo, nll, kl_alpha, kl_eta, kl_theta

    def init_hidden(self):
        """Initializes the first hidden state of the RNN used as inference network for \eta.
        """
        weight = next(self.parameters())
        nlayers = self.eta_nlayers
        nhid = self.eta_hidden_size
        return (weight.new_zeros(nlayers, 1, nhid), weight.new_zeros(nlayers, 1, nhid))



import os
import sys
from pathlib import Path

sys.path.append(os.getcwd())
octispath = Path(os.path.dirname(__file__)).parent.absolute().parent.absolute()
sys.path.append(str(octispath))

import argparse
import webbrowser
import frameworkScanner as fs
import octis.configuration.defaults as defaults
from multiprocessing import Process, Pool
import json
from flask import Flask, render_template, request, send_file
import tkinter as tk
import pandas as pd
import numpy as np
from tkinter import filedialog


app = Flask(__name__)
queueManager = ""


@app.route('/downloadSingleExp',
           methods=['GET'])
def downloadSingleExp():
    """
    Download the report of the given experiment

    :erturn: file with the report
    :rtype: File
    """
    experimentId = request.args.get("experimentId")
    batchId = request.args.get("batchId")

    paused = False
    if (experimentId + batchId) == queueManager.running:
        paused = True
        queueManager.pause()

    expPath = ""
    createdPath = os.path.join(
        queueManager.getExperiment(batchId, experimentId)["path"],
        experimentId,
        experimentId)
    jsonReport = {}
    if os.path.isfile(createdPath+".json"):
        expPath = createdPath

    with open(expPath+".json") as p:
        jsonReport = json.load(p)

    info = queueManager.getExperimentInfo(batchId, experimentId)

    n_row = info["current_iteration"]
    n_extra_metrics = len(jsonReport["extra_metric_names"])

    df = pd.DataFrame()
    df['dataset'] = [jsonReport["dataset_name"]] * n_row
    df['surrogate model'] = [jsonReport["surrogate_model"]] * n_row
    df['acquisition function'] = [jsonReport["acq_func"]] * n_row
    df['num_iteration'] = [i for i in range(n_row)]
    df['time'] = [jsonReport['time_eval'][i] for i in range(n_row)]
    df['Median(model_runs)'] = [np.median(
        jsonReport['dict_model_runs'][jsonReport['metric_name']]['iteration_' + str(i)]) for i in range(n_row)]
    df['Mean(model_runs)'] = [np.mean(
        jsonReport['dict_model_runs'][jsonReport['metric_name']]['iteration_' + str(i)]) for i in range(n_row)]
    df['Standard_Deviation(model_runs)'] = [np.std(
        jsonReport['dict_model_runs'][jsonReport['metric_name']]['iteration_' + str(i)]) for i in range(n_row)]

    for hyperparameter in list(jsonReport["x_iters"]):
        df[hyperparameter] = jsonReport["x_iters"][hyperparameter][0:n_row]

    for metric, i in zip(jsonReport["extra_metric_names"], range(n_extra_metrics)):
        df[metric + '(median, not optimized)'] = [np.median(
            jsonReport["dict_model_runs"][metric]['iteration_' + str(i)]) for i in range(n_row)]

        df[metric + '(Mean, not optimized)'] = [np.mean(
            jsonReport["dict_model_runs"][metric]['iteration_' + str(i)]) for i in range(n_row)]

        df[metric + '(Standard_Deviation, not optimized)'] = [np.std(
            jsonReport["dict_model_runs"][metric]['iteration_' + str(i)]) for i in range(n_row)]

    name_file = expPath + ".csv"

    df.to_csv(name_file, index=False, na_rep='Unkown')

    if paused:
        queueManager.start()

    return send_file(expPath+".csv",
                     mimetype="text/csv",
                     attachment_filename="report.csv",
                     as_attachment=True)


@ app.route("/selectPath", methods=['POST'])
def selectPath():
    """
    Select a path from the server and return it to the page

    :return: path
    :rtype: Dict
    """
    window = tk.Tk()
    path = filedialog.askdirectory()
    window.destroy()
    return {"path": path}


@ app.route("/serverClosed")
def serverClosed():
    """
    Reroute to the serverClosed page before server shutdown

    :return: template
    :rtype: render template
    """
    return render_template("serverClosed.html")


@ app.route("/shutdown")
def shutdown():
    """
    Save the state of the QueueManager and perform server shutdown

    :return: Ack signal
    :rtype: Dict
    """
    queueManager.stop()
    shutdown_server()
    return {"DONE": "YES"}


@ app.route('/')
def home():
    """
    Return the octis landing page

    :return: template
    :rtype: render template
    """
    return render_template("index.html")


@ app.route('/startExperiment', methods=['POST'])
def startExperiment():
    """
    Add a new experiment to the queue

    :return: template
    :rtype: render template
    """
    data = request.form.to_dict(flat=False)
    batch = data["batchId"][0]
    experimentId = data["expId"][0]
    if queueManager.getExperiment(batch, experimentId):
        return VisualizeExperiments()

    expParams = dict()
    expParams["partitioning"] = ("partitioning" in data)
    expParams["path"] = data["path"][0]
    expParams["dataset"] = data["dataset"][0]
    expParams["model"] = {"name": data["model"][0]}
    expParams["optimization"] = {
        "iterations": typed(data["iterations"][0]),
        "model_runs": typed(data["runs"][0]),
        "surrogate_model": data["surrogateModel"][0],
        "n_random_starts": typed(data["n_random_starts"][0]),
        "acquisition_function": data["acquisitionFunction"][0],
        "search_spaces": {}
    }
    expParams["optimize_metrics"] = []
    expParams["track_metrics"] = []

    model_parameters_to_optimize = []

    for key, value in data.items():
        if "_check" in key:
            model_parameters_to_optimize.append(key.replace("_check", ''))

    for key, value in data.items():
        if "model." in key:
            if any(par in key for par in model_parameters_to_optimize):
                if "_xminx" in key:
                    name = key.replace("_xminx", '').replace("model.", '')
                    if name not in expParams["optimization"]["search_spaces"]:
                        expParams["optimization"]["search_spaces"][name] = {}
                    expParams["optimization"]["search_spaces"][name]["low"] = typed(
                        value[0])
                elif "_xmaxx" in key:
                    name = key.replace("_xmaxx", '').replace("model.", '')
                    if name not in expParams["optimization"]["search_spaces"]:
                        expParams["optimization"]["search_spaces"][name] = {}
                    expParams["optimization"]["search_spaces"][name]["high"] = typed(
                        value[0])
                elif "_check" not in key:
                    expParams["optimization"]["search_spaces"][key.replace(
                        "model.", '')] = request.form.getlist(key)
            else:
                if "name" in key:
                    expParams["model"][key.replace("model.", '')] = value[0]
                else:
                    if "parameters" not in expParams["model"]:
                        expParams["model"]["parameters"] = {}
                    expParams["model"]["parameters"][key.replace(
                        "model.", '')] = typed(value[0])

        if "metric." in key:
            optimize = True
            metric = {"name": key.replace("metric.", ''), "parameters": {}}
            for singleValue in value:

                for key, content in json.loads(singleValue).items():
                    if key != "metric" and key != "type":
                        metric["parameters"][key] = typed(content)

                    if key == "type" and content == "track":
                        optimize = False
                if optimize:
                    expParams["optimize_metrics"].append(metric)
                else:
                    expParams["track_metrics"].append(metric)



    if expParams["optimize_metrics"][0]["name"]=="F1Score" and not expParams["partitioning"]:
        return VisualizeExperiments()
    for trackedMetric in expParams["track_metrics"]:
        if trackedMetric["name"] == "F1Score" and not expParams["partitioning"]:
            return VisualizeExperiments()

    print(expParams)

    queueManager.add_experiment(batch, experimentId, expParams)
    return CreateExperiments()


@ app.route("/getBatchExperiments", methods=['POST'])
def getBatchExperiments():
    """
    return the information related to the experiments of a batch

    :return: informations of the experiment
    :rtype: Dict
    """
    data = request.json['data']
    experiments = []
    for key in data:
        batch_experiments = queueManager.getBatchExperiments(key)
        for experiment in batch_experiments:
            new_exp = experiment
            new_exp["optimization_data"] = queueManager.getExperimentInfo(
                experiment["batchId"],
                experiment["experimentId"])
            experiments.append(new_exp)
    return json.dumps(experiments)


@ app.route('/CreateExperiments')
def CreateExperiments():
    """
    Serve the experiment creation page

    :return: template
    :rtype: render template
    """
    models = defaults.model_hyperparameters
    models_descriptions = defaults.model_descriptions
    datasets = fs.scanDatasets()
    metrics = defaults.metric_parameters
    optimization = defaults.optimization_parameters
    return render_template("CreateExperiments.html",  datasets=datasets, models=models, metrics=metrics,
                           optimization=optimization, models_descriptions=models_descriptions)


@ app.route('/VisualizeExperiments')
def VisualizeExperiments():
    """
    Serve the experiments visualization page

    :return: template
    :rtype: render template
    """
    batch_names = queueManager.getBatchNames()
    return render_template("VisualizeExperiments.html",
                           batchNames=batch_names)


@ app.route('/ManageExperiments')
def ManageExperiments():
    """
    Serve the ManageExperiments page

    :return: template
    :rtype: render template
    """
    exp_list = queueManager.getToRun()
    for exp in exp_list:
        exp_info = queueManager.getExperimentInfo(
            exp_list[exp]["batchId"], exp_list[exp]["experimentId"])
        if exp_info is not None:
            exp_list[exp].update(exp_info)
    order = queueManager.getOrder()
    running = queueManager.getRunning()
    return render_template("ManageExperiments.html", order=order, experiments=exp_list, running=running)


@ app.route("/pauseExp", methods=["POST"])
def pauseExp():
    """
    Pause the current experiment

    :return: ack signal
    :rtype: Dict
    """
    queueManager.pause()
    return {"DONE": "YES"}


@ app.route("/startExp", methods=["POST"])
def startExp():
    """
    Start the next experiment in the queue

    :return: ack signal
    :rtype: Dict
    """
    print(queueManager.getRunning())
    if queueManager.getRunning() == None:
        queueManager.next()
    return {"DONE": "YES"}


@ app.route("/deleteExp", methods=["POST"])
def deleteExp():
    """
    Delete the selected experiment from the queue

    :return: ack signal
    :rtype: Dict
    """
    data = request.json['data']
    print(queueManager.getRunning())
    if queueManager.getRunning() is not None and queueManager.getRunning() == data:
        queueManager.pause()
        queueManager.deleteFromOrder(data)
    else:
        queueManager.deleteFromOrder(data)
    return {"DONE": "YES"}


@ app.route("/updateOrder", methods=["POST"])
def updateOrder():
    """
    Update the order of the experiments in the queue

    :return: ack signal
    :rtype: Dict
    """
    data = request.json['data']
    queueManager.editOrder(data)
    return {"DONE": "YES"}


@ app.route("/getDocPreview", methods=["POST"])
def getDocPreview():
    """
    Returns the first 40 words of the selected document

    :return: first 40 words of the document
    :rtype: Dict
    """
    data = request.json['data']
    return json.dumps({"doc": fs.getDocPreview(data["dataset"], int(data["document"]))})


@ app.route('/SingleExperiment/<batch>/<exp_id>')
def SingleExperiment(batch="", exp_id=""):
    """
    Serve the single experiment page

    :return: template
    :rtype: render template
    """
    models = defaults.model_hyperparameters
    output = queueManager.getModel(batch, exp_id, 0, 0)
    global_info = queueManager.getExperimentInfo(batch, exp_id)
    iter_info = queueManager.getExperimentIterationInfo(batch, exp_id, 0)
    exp_info = queueManager.getExperiment(batch, exp_id)
    exp_ids = queueManager.getAllExpIds()
    vocabulary_path = os.path.join(exp_info["path"],
                                   exp_info["experimentId"],
                                   "models",
                                   "vocabulary.json")
    vocabulary = fs.getVocabulary(vocabulary_path)

    return render_template("SingleExperiment.html", batchName=batch, experimentName=exp_id,
                           output=output, globalInfo=global_info, iterationInfo=iter_info,
                           expInfo=exp_info, expIds=exp_ids, datasetMetadata=fs.getDatasetMetadata(
                               exp_info["dataset"]), vocabulary=vocabulary, models=models)


@ app.route("/getIterationData", methods=["POST"])
def getIterationData():
    """
    Return data of a single iteration and model run of an experiment

    :return: data of a single iteration and model run of an experiment
    :rtype: Dict
    """
    data = request.json['data']
    output = queueManager.getModel(data["batchId"], data["experimentId"],
                                   int(data["iteration"]), data["model_run"])
    iter_info = queueManager.getExperimentIterationInfo(data["batchId"], data["experimentId"],
                                                        int(data["iteration"]))
    return {"iterInfo": iter_info, "output": output}


def typed(value):
    """
    Handles typing of data

    :param value: value to cast
    :type value: *

    :raises ValueError: cannot cast data

    :return: data with the right type
    :rtype: *
    """
    try:
        t = int(value)
        return t
    except ValueError:
        try:
            t = float(value)
            return t
        except ValueError:
            return value


def shutdown_server():
    """
    Perform server shutdown

    :raise RuntimeError: wrong server environment used
    """
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()


if __name__ == '__main__':
    """
    Initialize the server
    """
    from queueManager import QueueManager

    
    parser = argparse.ArgumentParser()
    parser.add_argument("--port", type=int, help="port", default=5000)
    parser.add_argument("--host", type=str, help="host", default='localhost')
    parser.add_argument("--dashboardState", type=str, help="dashboardState", default="")

    args = parser.parse_args()

    dashboardState = None
    if args.dashboardState != "":
        dashboardState = args.dashboardState
    else:
        dashboardState = os.path.join(os.getcwd(),"queueManagerState.json")

    queueManager = QueueManager(dashboardState)

    url = 'http://' + str(args.host) + ':' + str(args.port)
    webbrowser.open_new(url)
    app.run(port=args.port)



import numpy as np
import torch

def get_batch(tokens, counts, ind, vocab_size, device):
    """fetch input data by batch."""
    batch_size = len(ind)
    data_batch = np.zeros((batch_size, vocab_size))
    for i, doc_id in enumerate(ind):
        doc = tokens[doc_id]
        count = counts[doc_id]
        if doc_id != -1:
            for j, word in enumerate(doc):
                data_batch[i, word] = count[j]
    data_batch = torch.from_numpy(data_batch).float().to(device)
    return data_batch



from __future__ import division

from bisect import bisect_left
from collections import namedtuple
import numpy as np
from scipy.spatial import distance
import math


RBO = namedtuple("RBO", "min res ext")
RBO.__doc__ += ": Result of full RBO analysis"
RBO.min.__doc__ = "Lower bound estimate"
RBO.res.__doc__ = "Residual corresponding to min; min + res is an upper bound estimate"
RBO.ext.__doc__ = "Extrapolated point estimate"


def logistic(x, k):
    return 1/ (1 + math.exp(-k*x))


def _round(obj):
    if isinstance(obj, RBO):
        return RBO(_round(obj.min), _round(obj.res), _round(obj.ext))
    else:
        return round(obj, 3)


def set_at_depth(lst, depth):
    ans = set()
    for v in lst[:depth]:
        if isinstance(v, set):
            ans.update(v)
        else:
            ans.add(v)
    return ans


def embeddings_overlap(list1, list2, depth, index2word, embedding_space, norm=True):
    set1, set2 = set_at_depth(list1, depth), set_at_depth(list2, depth)
    word_list1 = [index2word[index] for index in list1]
    word_list2 = [index2word[index] for index in list2]

    centroid_1 = np.mean([embedding_space[w] for w in word_list1[:depth] if w in embedding_space], axis=0)
    centroid_2 = np.mean([embedding_space[w] for w in word_list2[:depth] if w in embedding_space], axis=0)
    cos_sim = 1 - distance.cosine(centroid_1, centroid_2)
    if cos_sim > 1:
        cos_sim = 1
    elif cos_sim < -1:
        cos_sim = -1
    if norm:
        e_ov = 1 - (math.acos(cos_sim) / math.pi)
    else:
        e_ov = cos_sim
    return e_ov, len(set1), len(set2)


def overlap(list1, list2, depth, index2word, embedding_space, norm):
    # return agreement(list1, list2, depth) * min(depth, len(list1), len(list2))
    # NOTE: comment the preceding and uncomment the following line if you want
    # to stick to the algorithm as defined by the paper
    ov = embeddings_overlap(list1, list2, depth, index2word, embedding_space, norm=norm)[0]
    # print("overlap", ov)
    return ov


def agreement(list1, list2, depth, index2word, embedding_space, norm):
    """Proportion of shared values between two sorted lists at given depth."""
    len_intersection, len_set1, len_set2 = embeddings_overlap(list1, list2, depth, index2word, embedding_space=embedding_space, norm=norm)
    return 2 * len_intersection / (len_set1 + len_set2)


def cumulative_agreement(list1, list2, depth, index2word, embedding_space, norm):
    return (agreement(list1, list2, d, index2word, embedding_space, norm) for d in range(1, depth + 1))


def rbo_min(list1, list2, p, index2word, embedding_space, norm=True, depth=None):
    """Tight lower bound on RBO.
    See equation (11) in paper.
    """
    depth = min(len(list1), len(list2)) if depth is None else depth
    x_k = overlap(list1, list2, depth, index2word, embedding_space, norm)
    log_term = x_k * math.log(1 - p)
    sum_term = sum(
        p ** d / d * (overlap(list1, list2, d, index2word, embedding_space=embedding_space, norm=norm) - x_k) for d in range(1, depth + 1)
    )
    return (1 - p) / p * (sum_term - log_term)


def rbo_res(list1, list2, p, index2word, embedding_space, norm=True):
    """Upper bound on residual overlap beyond evaluated depth.
    See equation (30) in paper.
    NOTE: The doctests weren't verified against manual computations but seem
    plausible. In particular, for identical lists, ``rbo_min()`` and
    ``rbo_res()`` should add up to 1, which is the case.
    """
    S, L = sorted((list1, list2), key=len)
    s, l = len(S), len(L)
    x_l = overlap(list1, list2, l, index2word, embedding_space, norm=norm)
    # since overlap(...) can be fractional in the general case of ties and f
    # must be an integer --> math.ceil()
    f = int(math.ceil(l + s - x_l))
    # upper bound of range() is non-inclusive, therefore + 1 is needed
    term1 = s * sum(p ** d / d for d in range(s + 1, f + 1))
    term2 = l * sum(p ** d / d for d in range(l + 1, f + 1))
    term3 = x_l * (math.log(1 / (1 - p)) - sum(p ** d / d for d in range(1, f + 1)))
    return p ** s + p ** l - p ** f - (1 - p) / p * (term1 + term2 + term3)


def rbo_ext(list1, list2, p, index2word, embedding_space, norm=True):
    """RBO point estimate based on extrapolating observed overlap.
    See equation (32) in paper.
    NOTE: The doctests weren't verified against manual computations but seem
    plausible.
    """
    S, L = sorted((list1, list2), key=len)
    s, l = len(S), len(L)
    x_l = overlap(list1, list2, l, index2word, embedding_space=embedding_space, norm=norm)
    x_s = overlap(list1, list2, s, index2word, embedding_space=embedding_space, norm=norm)
    # the paper says overlap(..., d) / d, but it should be replaced by
    # agreement(..., d) defined as per equation (28) so that ties are handled
    # properly (otherwise values > 1 will be returned)
    # sum1 = sum(p**d * overlap(list1, list2, d)[0] / d for d in range(1, l + 1))
    sum1 = sum(p ** d * agreement(list1, list2, d, index2word=index2word, embedding_space=embedding_space, norm=norm)
               for d in range(1, l + 1))
    sum2 = sum(p ** d * x_s * (d - s) / s / d for d in range(s + 1, l + 1))
    term1 = (1 - p) / p * (sum1 + sum2)
    term2 = p ** l * ((x_l - x_s) / l + x_s / s)
    return term1 + term2


def word_embeddings_rbo(list1, list2, p, index2word, embedding_space, norm):
    """Complete RBO analysis (lower bound, residual, point estimate).
    ``list`` arguments should be already correctly sorted iterables and each
    item should either be an atomic value or a set of values tied for that
    rank. ``p`` is the probability of looking for overlap at rank k + 1 after
    having examined rank k.
    RBO(min=0.489, res=0.477, ext=0.967)
    """
    if not 0 <= p <= 1:
        raise ValueError("The ``p`` parameter must be between 0 and 1.")
    args = (list1, list2, p, index2word, embedding_space, norm)

    return RBO(rbo_min(*args), rbo_res(*args), rbo_ext(*args))


def sort_dict(dct, *, ascending=False):
    """Sort keys in ``dct`` according to their corresponding values.
    Sorts in descending order by default, because the values are
    typically scores, i.e. the higher the better. Specify
    ``ascending=True`` if the values are ranks, or some sort of score
    where lower values are better.
    Ties are handled by creating sets of tied keys at the given position
    in the sorted list.
    >>> dct = dict(a=1, b=2, c=1, d=3)
    >>> list(sort_dict(dct)) == ['d', 'b', {'a', 'c'}]
    True
    >>> list(sort_dict(dct, ascending=True)) == [{'a', 'c'}, 'b', 'd']
    True
    """
    scores = []
    items = []
    # items should be unique, scores don't have to
    for item, score in dct.items():
        if not ascending:
            score *= -1
        i = bisect_left(scores, score)
        if i == len(scores):
            scores.append(score)
            items.append(item)
        elif scores[i] == score:
            existing_item = items[i]
            if isinstance(existing_item, set):
                existing_item.add(item)
            else:
                items[i] = {existing_item, item}
        else:
            scores.insert(i, score)
            items.insert(i, item)
    return items


def rbo_dict(dict1, dict2, p, index2word, embedding_space, norm, *, sort_ascending=False):
    """Wrapper around ``rbo()`` for dict input.
    Each dict maps items to be sorted to the score according to which
    they should be sorted. The RBO analysis is then performed on the
    resulting sorted lists.
    The sort is descending by default, because scores are typically the
    higher the better, but this can be overridden by specifying
    ``sort_ascending=True``.
    """
    list1, list2 = (
        sort_dict(dict1, ascending=sort_ascending),
        sort_dict(dict2, ascending=sort_ascending),
    )
    return word_embeddings_rbo(list1, list2, p, index2word, embedding_space, norm)



from octis.models.model import AbstractModel
import numpy as np
from sklearn.decomposition import NMF
from sklearn.feature_extraction.text import TfidfVectorizer
import octis.configuration.defaults as defaults


class NMF_scikit(AbstractModel):

    def __init__(
        self, num_topics=100, init=None, alpha=0, l1_ratio=0,
            regularization='both', use_partitions=True):
        """
        Initialize NMF model

        Parameters
        ----------
        num_topics (int) – Number of topics to extract.

        init (string, optional) – Method used to initialize the procedure.
        Default: None. Valid options:

            None: ‘nndsvd’ if n_components <= min(n_samples, n_features),
            otherwise random.

            ‘random’: non-negative random matrices, scaled with:
            sqrt(X.mean() / n_components)

            ‘nndsvd’: Nonnegative Double Singular Value Decomposition (NNDSVD)
            initialization (better for sparseness)

            ‘nndsvda’: NNDSVD with zeros filled with the average of X
            (better when sparsity is not desired)

            ‘nndsvdar’: NNDSVD with zeros filled with small random values
            (generally faster, less accurate alternative to NNDSVDa for when
            sparsity is not desired)

        alpha (double, optional) – Constant that multiplies the regularization
        terms. Set it to zero to have no regularization.

        l1_ratio (double, optional) – The regularization mixing parameter, with
        0 <= l1_ratio <= 1. For l1_ratio = 0 the penalty is an elementwise
        L2 penalty (aka Frobenius Norm). For l1_ratio = 1 it is an
        elementwise L1 penalty. For 0 < l1_ratio < 1, the penalty
        is a combination of L1 and L2.
        """
        super().__init__()
        self.hyperparameters["num_topics"] = num_topics
        self.hyperparameters["init"] = init
        self.hyperparameters["alpha"] = alpha
        self.hyperparameters["l1_ratio"] = l1_ratio
        self.hyperparameters['regularization'] = regularization
        self.use_partitions = use_partitions

        self.id2word = None
        self.id_corpus = None
        self.update_with_test = False

    def hyperparameters_info(self):
        """
        Returns hyperparameters informations
        """
        return defaults.NMF_scikit_hyperparameters_info

    def partitioning(self, use_partitions, update_with_test=False):
        """
        Handle the partitioning system to use and reset the model to perform
        new evaluations

        Parameters
        ----------
        use_partitions: True if train/set partitioning is needed, False
                        otherwise
        update_with_test: True if the model should be updated with the test set,
                          False otherwise
        """
        self.use_partitions = use_partitions
        self.update_with_test = update_with_test
        self.id2word = None
        self.id_corpus = None

    def train_model(self, dataset, hyperparameters=None, top_words=10):
        """
        Train the model and return output

        Parameters
        ----------
        dataset : dataset to use to build the model
        hyperparameters : hyperparameters to build the model
        top_words : if greather than 0 returns the most significant words
                 for each topic in the output
                 Default True


        Returns
        -------
        result : dictionary with up to 3 entries,
                 'topics', 'topic-word-matrix' and
                 'topic-document-matrix'
        """
        if hyperparameters is None:
            hyperparameters = {}

        if self.id2word is None or self.id_corpus is None:
            vectorizer = TfidfVectorizer(
                min_df=0.0, token_pattern=r"(?u)\b[\w|\-]+\b",
                vocabulary=dataset.get_vocabulary())

            if self.use_partitions:
                partition = dataset.get_partitioned_corpus(
                    use_validation=False)
                corpus = partition[0]
            else:
                corpus = dataset.get_corpus()

            real_corpus = [" ".join(document) for document in corpus]
            X = vectorizer.fit_transform(real_corpus)

            self.id2word = {i: k for i, k in enumerate(
                vectorizer.get_feature_names_out())}
            if self.use_partitions:
                test_corpus = []
                for document in partition[1]:
                    test_corpus.append(" ".join(document))
                Y = vectorizer.transform(test_corpus)
                self.id_corpus = X
                self.new_corpus = Y
            else:
                self.id_corpus = X

        #hyperparameters["corpus"] = self.id_corpus
        #hyperparameters["id2word"] = self.id2word
        self.hyperparameters.update(hyperparameters)
        model = NMF(
            n_components=self.hyperparameters["num_topics"],
            init=self.hyperparameters["init"],
            alpha=self.hyperparameters["alpha"],
            l1_ratio=self.hyperparameters["l1_ratio"],
            regularization=self.hyperparameters['regularization'])

        W = model.fit_transform(self.id_corpus)
        #W = W / W.sum(axis=1, keepdims=True)
        H = model.components_
        #H = H / H.sum(axis=1, keepdims=True)

        result = {}

        result["topic-word-matrix"] = H

        if top_words > 0:
            result["topics"] = self.get_topics(H, top_words)

        result["topic-document-matrix"] = np.array(W).transpose()

        if self.use_partitions:
            if self.update_with_test:
               # NOT IMPLEMENTED YET

                result["test-topic-word-matrix"] = W

                if top_words > 0:
                    result["test-topics"] = self.get_topics(W, top_words)

                result["test-topic-document-matrix"] = H

            else:
                result["test-topic-document-matrix"] = model.transform(
                    self.new_corpus).T

        return result

    def get_topics(self, H, top_words):
        topic_list = []
        for topic in H:
            words_list = sorted(
                list(enumerate(topic)), key=lambda x: x[1])
            topk = [tup[0] for tup in words_list[0:top_words]]
            topic_list.append([self.id2word[i] for i in topk])
        return topic_list



import numpy as np


def get_topic_diversity(beta, topk):
    num_topics = beta.shape[0]
    list_w = np.zeros((num_topics, topk))
    for k in range(num_topics):
        idx = beta[k, :].argsort()[-topk:][::-1]
        list_w[k, :] = idx
    n_unique = len(np.unique(list_w))
    TD = n_unique / (topk * num_topics)
    print('Topic diveristy is: {}'.format(TD))


def get_document_frequency(data, wi, wj=None):
    if wj is None:
        D_wi = 0
        for l in range(len(data)):
            doc = data[l].squeeze(0)
            if len(doc) == 1:
                continue
            else:
                doc = doc.squeeze()
            if wi in doc:
                D_wi += 1
        return D_wi
    D_wj = 0
    D_wi_wj = 0
    for l in range(len(data)):
        doc = data[l].squeeze(0)
        if len(doc) == 1:
            doc = [doc.squeeze()]
        else:
            doc = doc.squeeze()
        if wj in doc:
            D_wj += 1
            if wi in doc:
                D_wi_wj += 1
    return D_wj, D_wi_wj


def get_topic_coherence(beta, data, vocab):
    D = len(data)  # number of docs...data is list of documents
    print('D: ', D)
    TC = []
    num_topics = len(beta)
    for k in range(num_topics):
        print('k: {}/{}'.format(k, num_topics))
        top_10 = list(beta[k].argsort()[-11:][::-1])
        top_words = [vocab[a] for a in top_10]
        TC_k = 0
        counter = 0
        for i, word in enumerate(top_10):
            # get D(w_i)
            D_wi = get_document_frequency(data, word)
            j = i + 1
            tmp = 0
            while j < len(top_10) and j > i:
                # get D(w_j) and D(w_i, w_j)
                D_wj, D_wi_wj = get_document_frequency(data, word, top_10[j])
                # get f(w_i, w_j)
                if D_wi_wj == 0:
                    f_wi_wj = -1
                else:
                    f_wi_wj = -1 + (
                        np.log(D_wi) + np.log(D_wj) - 2.0 * np.log(D)) / (
                        np.log(D_wi_wj) - np.log(D))
                # update tmp:
                tmp += f_wi_wj
                j += 1
                counter += 1
            # update TC_k
            TC_k += tmp
        TC.append(TC_k)
    print('counter: ', counter)
    print('num topics: ', len(TC))
    TC = np.mean(TC) / counter
    print('Topic coherence is: {}'.format(TC))


def nearest_neighbors(word, embeddings, vocab):
    vectors = embeddings.data.cpu().numpy()
    index = vocab.index(word)
    print('vectors: ', vectors.shape)
    query = vectors[index]
    print('query: ', query.shape)
    ranks = vectors.dot(query).squeeze()
    denom = query.T.dot(query).squeeze()
    denom = denom * np.sum(vectors**2, 1)
    denom = np.sqrt(denom)
    ranks = ranks / denom
    mostSimilar = []
    [mostSimilar.append(idx) for idx in ranks.argsort()[::-1]]
    nearest_neighbors = mostSimilar[:20]
    nearest_neighbors = [vocab[comp] for comp in nearest_neighbors]
    return nearest_neighbors



import os
import time
import json
from collections import namedtuple
from experimentManager import startExperiment
import experimentManager as expManager
import multiprocessing as mp
from subprocess import Popen
import signal


class QueueManager:
    """
    The QueueManager class is used to track old, ongoing and new experiments
    """
    running = None
    toRun = None
    order = None
    completed = None
    process = None
    busy = None
    idle = None
    path = None

    def __init__(self, path):
        """
        Initialize the queue manager.
        Loads old queues
        """
        manager = mp.Manager()
        self.path = path
        self.running = manager.list()
        self.running.append(None)
        self.toRun = manager.dict()
        self.order = manager.list()
        self.completed = manager.dict()
        self.busy = manager.list()
        self.busy.append(False)
        self.process = manager.list()

        self.load_state(path)
        self.idle = mp.Process(target=self._run)
        self.idle.start()

    def save_state(self, path):
        """
        Saves the state of the queue
        """
        with open(path, "w") as fp:
            json.dump({"running": self.running[0],
                       "toRun": dict(self.toRun),
                       "order": list(self.order),
                       "completed": dict(self.completed)},
                      fp)

    def load_state(self, path):
        """
        Loads the state of the queue
        """
        if not os.path.isfile(path):
            with open(path, "w") as state:
                json.dump({"running": self.running[0],
                       "toRun": dict(self.toRun),
                       "order": list(self.order),
                       "completed": dict(self.completed)},
                       state)

        with open(path, "r") as fp:
            data = json.load(fp)
            self.running[0] = data["running"]
            self.toRun.update(data["toRun"])
            self.order.extend(data["order"])
            self.completed.update(data["completed"])

    def next(self):
        """
        If there is no running experiments, choose te next one to run

        :return: a tuple containing id of the batch and id of
                 the next experiment to run
        :rtype: tuple
        """
        if self.running[0] is None:
            self.running[0] = self.order.pop(0)
            self.busy[0] = False
            self.start()
        return self.running[0]

    def add_experiment(self, batch, id, parameters):
        """
        Adds a new experiment to the queue

        :param batch: id of the batch
        :type batch: String
        :param id: id of the experiment
        :type id: String
        :param parameters: dictionary with the parameters of the experiment
        :type parameters: Dict

        :return: True if the experiment was added to the queue, False otherwise
        :rtype: boolean
        """
        toAdd = batch+id
        parameters["batchId"] = batch
        parameters["experimentId"] = id
        if toAdd not in self.completed and toAdd not in self.toRun:
            self.toRun[toAdd] = parameters
            self.order.append(toAdd)
            return True
        return False

    def _run(self):
        """
        Put the current experiment in the finished queue

        :return: a tuple containing id of the batch and id of the
                 completed experiment
        :rtype: tuple
        """
        while True:
            time.sleep(4)
            if not self.busy[0]:
                if self.running[0] is not None:
                    finished = self.running[0]
                    self.completed[finished] = self.toRun[finished]
                    del self.toRun[finished]
                    self.running[0] = None
                    self.save_state(self.path)
                if len(self.order) > 0 and self.running[0] is None:
                    self.running[0] = self.order.pop(0)
                    self.start()
            time.sleep(6)

    def pause(self):
        """
        pause the running experiment

        :return: a tuple containing id of the batch and id of the
                 paused experiment
        :rtype: tuple
        """
        if self.busy[0] and self.running[0] != None:
            paused = self.running[0]
            to_stop = self.process.pop()
            os.kill(to_stop, signal.SIGTERM)
            self.order.insert(0, paused)
            self.running[0] = None
            return paused
        return False

    def getBatchNames(self):
        """
        Get the name of each batch with experiment in the completed list or
        in the list of experiments to run

        :return: names of each batch
        :rtype: List
        """
        batch_names = []
        to_remove = []
        for key, value in self.completed.items():
            if not self.getExperimentInfo(value["batchId"], value["experimentId"]):
                to_remove.append(key)
            else:
                if value["batchId"] not in batch_names:
                    batch_names.append(value["batchId"])
        for el in to_remove:
            del self.completed[el]

        for key, value in self.toRun.items():
            if value["batchId"] not in batch_names:
                batch_names.append(value["batchId"])
        return batch_names

    def getBatchExperiments(self, batch_name):
        """
        Retrieves all the experiments of the selected batch

        :param batch_name: name of the batch
        :type batch_name: String

        :return: list of experiments metadata
        :rtype: List
        """
        experiments = []
        to_remove = []
        for key, value in self.completed.items():
            if not self.getExperimentInfo(value["batchId"], value["experimentId"]):
                to_remove.append(key)
            else:
                if value["batchId"] == batch_name:
                    experiments.append(value)
        for el in to_remove:
            del self.completed[el]

        for key, value in self.toRun.items():
            if value["batchId"] == batch_name:
                experiments.append(value)
        return experiments

    def getExperimentInfo(self, batch, experimentId):
        """
        Return the info of the experiment with the given batch name and id

        :param batch: name of the batch
        :type batch: String
        :param experimentId: name of the experiment
        :type experimentId: String

        :return: experiment info (mean, median, best seen, worst seen)
        :rtype: Dict
        """
        experiment = None
        if batch + experimentId in self.completed:
            experiment = self.completed[batch+experimentId]
        if batch + experimentId in self.toRun:
            experiment = self.toRun[batch+experimentId]
        if experiment is not None:
            path = str(os.path.join(
                experiment["path"], experiment["experimentId"], experiment["experimentId"]+".json"))
            return expManager.singleInfo(path)
        return None

    def start(self):
        """
        Start a new experiment in a new process
        """
        if not self.busy[0]:
            self.busy[0] = True
            process = mp.Process(
                target=QueueManager._execute_and_update, args=(
                    self.toRun,
                    self.running,
                    self.busy
                ))
            process.start()
            print("starting "+self.running[0])
            self.process.append(process.pid)

    def stop(self):
        """
        Stop the current experiment and save the information about it
        """
        self.idle.terminate()
        self.pause()
        self.save_state(self.path)

    @staticmethod
    def _execute_and_update(toRun, running, busy):
        """
        start an experiment using a static method

        :param toRun: list of experiments to run
        :type: List
        :param running: id of the running experiment
        :type: String
        :param busy: specify if the queuemanager is busy
        :type busy: boolean
        """
        startExperiment(toRun[running[0]])
        busy[0] = False

    def getModel(self, batch, experimentId, iteration, modelRun):
        """
        Retrieve output of the model for a single model

        :param batch: name of the batch
        :type batch: String
        :param experimentId: name of the experiment
        :type experimentId: String
        :param iterarion: number of iteration of the model to retrieve
        :type iteration: Int
        :param modelRun: numeber of model run of the model to retrieve
        :type modelRun: Int

        :return: output of the model (topic-word-matrix,
                 document-topic-matrix and vocabulary)
        :rtype: Dict
        """
        experiment = None
        if batch+experimentId in self.completed:
            experiment = self.completed[batch+experimentId]
        if batch+experimentId in self.toRun:
            experiment = self.toRun[batch+experimentId]
        if experiment is not None:
            path = str(os.path.join(
                experiment["path"], experiment["experimentId"]))
            return expManager.getModelInfo(path, iteration, modelRun)
        return False

    def getExperimentIterationInfo(self, batch, experimentId, iteration=0):
        """
        Retrieve the results of the BO untile the given iteration

        :param batch: id of the batch
        :type batch: String
        :param experimentId: id of the experiment
        :type experimentId: String
        :param iteration: last iteration to consider
        :type iteration: Int

        :return: experiment interation info
        :rtype: Dict
        """
        experiment = None
        if batch+experimentId in self.completed:
            experiment = self.completed[batch+experimentId]
        if batch+experimentId in self.toRun:
            experiment = self.toRun[batch+experimentId]
        if experiment is not None:
            path = str(os.path.join(
                experiment["path"], experiment["experimentId"], experiment["experimentId"]+".json"))
            return expManager.retrieveIterationBoResults(path, iteration)
        return False

    def getExperiment(self, batch, experimentId):
        """
        Retrieve metadata about the experiment

        :param batch: name of the batch
        :type batch: String
        :param experiment: name of the experiment
        :type experiment: String

        :return: return experimen metadatata
        :rtype: Dict
        """
        experiment = False
        if batch+experimentId in self.completed:
            experiment = self.completed[batch+experimentId]
        if batch+experimentId in self.toRun:
            experiment = self.toRun[batch+experimentId]
        return experiment

    def getAllExpIds(self):
        """
        Retrieve the name of each experiment and their batch

        :param expIds: list of entries. Each entry is a list with 2 elements.
                 the name of the experiment and a list with name of the batch and
                 name of the experiment
        :type expIds: List

        :return: name and batch of each experiment
        :rtype: List
        """
        expIds = []
        to_remove = []
        for key, exp in self.completed.items():
            if not self.getExperimentInfo(exp["batchId"], exp["experimentId"]):
                to_remove.append(key)
            else:
                expIds.append([exp["experimentId"],
                               [exp["batchId"], exp["experimentId"]]])
        for el in to_remove:
            del self.completed[el]
        for key, exp in self.toRun.items():
            expIds.append([exp["experimentId"],
                           [exp["batchId"], exp["experimentId"]]])
        return expIds

    def getToRun(self):
        """
        Retrieve the experiments to run

        :return: dictionary of the experiments to run
        :rtype: Dict
        """
        return dict(self.toRun)

    def getOrder(self):
        """
        Retrieve the order of the experiments to run

        :return: id of each experiment to run in order (from first to last)
        :rtype: List
        """
        return list(self.order)

    def getRunning(self):
        """
        returns the id of the running experiment

        :return: id of the running experiment
        :rtype: String
        """
        return self.running[0]

    def editOrder(self, newOrder):
        """
        Updates the order of the experiments to run

        :param newOrder: new order of the experiments
        :type newOrder: List
        """
        FinalOrder = []
        for el in newOrder:
            if el in self.order:
                FinalOrder.append(el)
        self.order[:] = []
        self.order.extend(FinalOrder)

    def deleteFromOrder(self, experimentId):
        """
        Delete and experiment from the queue of experiments to run

        :param experimentId: id of the experiment to remove from the queue
        :type experimentId: String
        """
        self.order = list(filter(lambda a: a != experimentId, self.order))
        del self.toRun[experimentId]



import numpy as np
from sentence_transformers import SentenceTransformer
import scipy.sparse
import warnings
from octis.models.contextualized_topic_models.datasets.dataset import CTMDataset
import os
import pickle as pkl

def get_bag_of_words(data, min_length):
    """
    Creates the bag of words
    """
    vect = [np.bincount(x[x != np.array(None)].astype('int'), minlength=min_length)
            for x in data if np.sum(x[x != np.array(None)]) != 0]

    vect = scipy.sparse.csr_matrix(vect)
    return vect

def bert_embeddings_from_file(text_file, sbert_model_to_load, batch_size=200):
    """
    Creates SBERT Embeddings from an input file
    """
    model = SentenceTransformer(sbert_model_to_load)
    with open(text_file, encoding="utf-8") as filino:
        train_text = list(map(lambda x: x, filino.readlines()))

    return np.array(model.encode(train_text, show_progress_bar=True, batch_size=batch_size))


def bert_embeddings_from_list(texts, sbert_model_to_load="bert-base-nli-mean-tokens", batch_size=100):
    """
    Creates SBERT Embeddings from a list
    """
    model = SentenceTransformer(sbert_model_to_load)
    return np.array(model.encode(texts, show_progress_bar=True, batch_size=batch_size))


class QuickText:
    """
    Integrated class to handle all the text preprocessing needed
    """
    def __init__(self, bert_model, text_for_bow, text_for_bert=None, bert_path=None):
        """
        :param bert_model: string, bert model to use
        :param text_for_bert: list, list of sentences with the unpreprocessed text
        :param text_for_bow: list, list of sentences with the preprocessed text
        """
        self.vocab_dict = {}
        self.vocab = []
        self.index_dd = None
        self.idx2token = None
        self.bow = None
        self.bert_model = bert_model
        self.text_handler = ""
        self.data_bert = None
        self.text_for_bow = text_for_bow

        if text_for_bert is not None:
            self.text_for_bert = text_for_bert
        else:
            self.text_for_bert = None
        self.bert_path = bert_path

    def prepare_bow(self):
        indptr = [0]
        indices = []
        data = []
        vocabulary = {}

        if self.text_for_bow is not None:
            docs = self.text_for_bow
        else:
            docs = self.text_for_bert

        for d in docs:
            for term in d.split():
                index = vocabulary.setdefault(term, len(vocabulary))
                indices.append(index)
                data.append(1)
            indptr.append(len(indices))

        self.vocab_dict = vocabulary
        self.vocab = list(vocabulary.keys())

        warnings.simplefilter('always', DeprecationWarning)
        if len(self.vocab) > 2000:
            warnings.warn("The vocab you are using has more than 2000 words, reconstructing high-dimensional vectors requires"
                          "significantly more training epochs and training samples. "
                          "Consider reducing the number of vocabulary items. "
                          "See https://github.com/MilaNLProc/contextualized-topic-models#preprocessing "
                          "and https://github.com/MilaNLProc/contextualized-topic-models#tldr", Warning)

        self.idx2token = {v: k for (k, v) in self.vocab_dict.items()}
        self.bow = scipy.sparse.csr_matrix((data, indices, indptr), dtype=int)

    def load_contextualized_embeddings(self, embeddings):
        self.data_bert = embeddings

    def load_dataset(self):
        self.prepare_bow()
        if self.bert_path is not None:
            if os.path.exists(self.bert_path):
                self.data_bert = pkl.load(open(self.bert_path, 'r'))
        else:
            if self.data_bert is None:
                if self.text_for_bert is not None:
                    self.data_bert = bert_embeddings_from_list(self.text_for_bert, self.bert_model)
                else:
                    self.data_bert = bert_embeddings_from_list(self.text_for_bow, self.bert_model)
                pkl.dump(self.data_bert, open(self.bert_path, 'w'))

        training_dataset = CTMDataset(self.bow, self.data_bert, self.idx2token)
        return training_dataset

class TextHandler:
    """
    Class used to handle the text preparation and the BagOfWord
    """
    def __init__(self, file_name=None, sentences=None):
        self.file_name = file_name
        self.sentences = sentences
        self.vocab_dict = {}
        self.vocab = []
        self.index_dd = None
        self.idx2token = None
        self.bow = None

        warnings.simplefilter('always', DeprecationWarning)
        if len(self.vocab) > 2000:
            warnings.warn("TextHandler class is deprecated and will be removed in version 2.0. Use QuickText.", Warning)

    def prepare(self):
        indptr = [0]
        indices = []
        data = []
        vocabulary = {}

        if self.sentences is None and self.file_name is None:
            raise Exception("Sentences and file_names cannot both be none")

        if self.sentences is not None:
            docs = self.sentences
        elif self.file_name is not None:
            with open(self.file_name, encoding="utf-8") as filino:
                docs = filino.readlines()
        else:
            raise Exception("One parameter between sentences and file_name should be selected")

        for d in docs:
            for term in d.split():
                index = vocabulary.setdefault(term, len(vocabulary))
                indices.append(index)
                data.append(1)
            indptr.append(len(indices))

        self.vocab_dict = vocabulary
        self.vocab = list(vocabulary.keys())

        warnings.simplefilter('always', DeprecationWarning)
        if len(self.vocab) > 2000:
            warnings.warn("The vocab you are using has more than 2000 words, reconstructing high-dimensional vectors requires"
                          "significantly more training epochs and training samples. "
                          "Consider reducing the number of vocabulary items. "
                          "See https://github.com/MilaNLProc/contextualized-topic-models#preprocessing "
                          "and https://github.com/MilaNLProc/contextualized-topic-models#tldr", Warning)

        self.idx2token = {v: k for (k, v) in self.vocab_dict.items()}
        self.bow = scipy.sparse.csr_matrix((data, indices, indptr), dtype=int)



import os
import random
import pickle
import pickle
import numpy as np
import torch 
import scipy.io
from pathlib import Path
from gensim.models.fasttext import FastText as FT_gensim
from tqdm import tqdm
from scipy.io import loadmat
from tqdm import tqdm
from sklearn.model_selection import train_test_split
from scipy.sparse import coo_matrix, hstack

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def read_mat_file(key, path):
    """
    read the preprocess mat file whose key and and path are passed as parameters

    Args:
        key ([type]): [description]
        path ([type]): [description]

    Returns:
        [type]: [description]
    """
    term_path = Path().cwd().joinpath('data', 'preprocess', path)
    doc = loadmat(term_path)[key]
    return doc


def get_time_columns(matrix):
    """
    given a matrix where the time column is the last row return the doc term matrix and the times
    """
    doc_term_matrix = matrix[:, :-1]
    times = matrix[:, -1].toarray().astype(np.int32).squeeze()
    return doc_term_matrix, times


def split_train_test_matrix(dataset):
    """Split the dataset into the train set , the validation and the test set

    Args:
        dataset ([type]): [description]

    Returns:
        [type]: [description]
    """
    X_train, X_test_global = train_test_split(dataset, test_size=0.25, random_state=1)
    X_val, X_test = train_test_split(X_test_global, test_size=0.5, random_state=1)
    X_test_1, X_test_2 = train_test_split(X_test, test_size=0.5, random_state=1)
    X_test_val, X_test = train_test_split(X_val, test_size=0.5, random_state=1)
    return X_train, X_val, X_test_1, X_test_2, X_test


def get_data(doc_terms_file_name="tf_idf_doc_terms_matrix",
             terms_filename="tf_idf_terms",
             timestamp_filename='tf_idf_timestamps'):
    """read the data and return the vocabulary as well as the train, test and validation tests

    Args:
        path ([type]): [description]

    Returns:
        [type]: [description]
    """
    doc_term_matrix = read_mat_file("doc_terms_matrix", doc_terms_file_name)
    terms = read_mat_file("terms", terms_filename)
    times = read_mat_file("timestamps", timestamp_filename)
    vocab = terms
    #  make the time the last row of the doc_terms matrix for better spliting
    times = times.reshape((times.shape[1], 1))
    doc_term_time_matrix = hstack([doc_term_matrix, times])
    train, validation, test_1, test_2, test = split_train_test_matrix(doc_term_time_matrix)
    return vocab, train, validation, test_1, test_2, test


def get_batch(doc_terms_matrix, indices, device, timestamps):
    """
    get the a sample of the given indices

    Basically get the given indices from the dataset

    Args:
        doc_terms_matrix ([type]): the document term matrix
        indices ([type]):  numpy array
        vocab_size ([type]): [description]

    Returns:
        [numpy array]: a numpy array with the data passed as parameter
    """
    data_batch = doc_terms_matrix[indices, :]
    data_batch = torch.from_numpy(data_batch.toarray()).float().to(device)
    times_batch = timestamps[indices]
    return data_batch, times_batch


def train_or_load(args1, args2):
    """This decorator load the data passed in parameter or return the train function

    Args:
        func ([type]): [description]
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            if args[3] in ["train", "test", "valid", "test_1", "test_2"]:
                data_path = args1.joinpath(f"rnn_input_{args[3]}.pkl")
                kwargs['data_path'] = data_path
            if args2:
                dataset = torch.load(data_path)
                return dataset
            return func(*args, **kwargs)
        return wrapper
    return decorator


@train_or_load(Path.cwd().joinpath("data", "rnn_inputs"), True)
def get_rnn_input(doc_terms_matrix, num_times, vocab_size, mode, **kwargs):
    """
    need to understand this part

    Args:
        doc_terms_matrix ([type]): [description]
        counts ([type]): [description]
        timestamps ([type]): [description]
        num_times ([type]): [description]
        vocab_size ([type]): [description]
        num_docs ([type]): [description]

    Returns:
        [type]: [description]
    """
    doc_terms_matrix, timestamps = get_time_columns(doc_terms_matrix)
    num_docs = doc_terms_matrix.shape[0]
    indices = torch.randperm(num_docs)
    indices = torch.split(indices, 1000)
    rnn_input = torch.zeros(num_times, vocab_size).to(device)
    cnt = torch.zeros(num_times, ).to(device)
    for idx, ind in tqdm(enumerate(indices)):
        data_batch, times_batch = get_batch(doc_terms_matrix, ind, device, timestamps)
        # this part I will look into it after
        for t in range(num_times):
            tmp = (times_batch == t).nonzero()
            docs = data_batch[tmp].squeeze().sum(0)
            rnn_input[t] += docs
            cnt[t] += len(tmp)
        if idx % 20 == 0:
            print('idx: {}/{}'.format(idx, len(indices)))
    rnn_input = rnn_input / cnt.unsqueeze(1)
    torch.save(rnn_input, kwargs.get("data_path"))
    return rnn_input


@train_or_load(Path().cwd().joinpath('data', 'preprocess', "embedding_matrix.npy"), True)
def read_embedding_matrix(vocab, device,  data_path):
    """
    read the embedding  matrix passed as parameter and return it as an vocabulary of each word
    with the corresponding embeddings

    Args:
        path ([type]): [description]

    # we need to use tensorflow embedding lookup heer
    """
    model_path = Path.cwd().joinpath("octis",
                                     "models",
                                     "ETM_model",
                                     "data",
                                     'embeddings_one_gram_fast_tweets_only').__str__()
    embeddings_path = Path().cwd().joinpath('data', 'preprocess', "embedding_matrix.npy")
    model_gensim = FT_gensim.load(model_path)
    vectorized_get_embeddings = np.vectorize(model_gensim.wv.get_vector)
    embeddings_matrix = np.zeros(shape=(len(vocab),50)) #should put the embeding size as a vector
    print("starting getting the word embeddings ++++ ")
    vocab = vocab.ravel()
    for index, word in tqdm(enumerate(vocab)):
        vector = model_gensim.wv.get_vector(word)
        embeddings_matrix[index] = vector
    print("done getting the word embeddings ")
    with open(embeddings_path, 'wb') as file_path:
        np.save(file_path, embeddings_matrix)
    embeddings = torch.from_numpy(embeddings_matrix).to(device)
    return embeddings



from octis.models.model import AbstractModel
import numpy as np
from gensim.models import ldamodel
import gensim.corpora as corpora
import octis.configuration.citations as citations
import octis.configuration.defaults as defaults


class LDA(AbstractModel):

    id2word = None
    id_corpus = None
    use_partitions = True
    update_with_test = False

    def __init__(
        self, num_topics=100, distributed=False, chunksize=2000,
        passes=1, update_every=1, alpha="symmetric", eta=None, decay=0.5,
        offset=1.0, eval_every=10, iterations=50, gamma_threshold=0.001,
            random_state=None):
        """
        Initialize LDA model

        Parameters
        ----------
        num_topics (int, optional) – The number of requested latent topics to
        be extracted from the training corpus.

        distributed (bool, optional) – Whether distributed computing should be
        used to accelerate training.

        chunksize (int, optional) – Number of documents to be used in each
        training chunk.

        passes (int, optional) – Number of passes through the corpus during
        training.

        update_every (int, optional) – Number of documents to be iterated
        through for each update. Set to 0 for batch learning, > 1 for
        online iterative learning.

        alpha ({numpy.ndarray, str}, optional) – Can be set to an 1D array of
        length equal to the number of expected topics that expresses our
        a-priori belief for the each topics’ probability. Alternatively
        default prior selecting strategies can be employed by supplying
        a string:

            ’asymmetric’: Uses a fixed normalized asymmetric prior of
            1.0 / topicno.

            ’auto’: Learns an asymmetric prior from the corpus
            (not available if distributed==True).

        eta ({float, np.array, str}, optional) – A-priori belief on word
        probability, this can be:

            scalar for a symmetric prior over topic/word probability,

            vector of length num_words to denote an asymmetric user defined
            probability for each word,

            matrix of shape (num_topics, num_words) to assign a probability
            for each word-topic combination,

            the string ‘auto’ to learn the asymmetric prior from the data.

        decay (float, optional) – A number between (0.5, 1] to weight what
        percentage of the previous lambda value is forgotten when each new
        document is examined.

        offset (float, optional) – Hyper-parameter that controls how much
        we will slow down the first steps the first few iterations.

        eval_every (int, optional) – Log perplexity is estimated every
        that many updates. Setting this to one slows down training by ~2x.

        iterations (int, optional) – Maximum number of iterations through the
        corpus when inferring the topic distribution of a corpus.

        gamma_threshold (float, optional) – Minimum change in the value of the
        gamma parameters to continue iterating.

        random_state ({np.random.RandomState, int}, optional) – Either a
        randomState object or a seed to generate one.s
        Useful for reproducibility.


        """
        super().__init__()
        self.hyperparameters = dict()
        self.hyperparameters["num_topics"] = num_topics
        self.hyperparameters["distributed"] = distributed
        self.hyperparameters["chunksize"] = chunksize
        self.hyperparameters["passes"] = passes
        self.hyperparameters["update_every"] = update_every
        self.hyperparameters["alpha"] = alpha
        self.hyperparameters["eta"] = eta
        self.hyperparameters["decay"] = decay
        self.hyperparameters["offset"] = offset
        self.hyperparameters["eval_every"] = eval_every
        self.hyperparameters["iterations"] = iterations
        self.hyperparameters["gamma_threshold"] = gamma_threshold
        self.hyperparameters["random_state"] = random_state

    def info(self):
        """
        Returns model informations
        """
        return {
            "citation": citations.models_LDA,
            "name": "LDA, Latent Dirichlet Allocation"
        }

    def hyperparameters_info(self):
        """
        Returns hyperparameters informations
        """
        return defaults.LDA_hyperparameters_info

    def set_hyperparameters(self, **kwargs):
        """
        Set model hyperparameters
        """
        super().set_hyperparameters(**kwargs)
        # Allow alpha to be a float in case of symmetric alpha
        if "alpha" in kwargs:
            if isinstance(kwargs["alpha"], float):
                self.hyperparameters["alpha"] = [
                    kwargs["alpha"]
                ] * self.hyperparameters["num_topics"]

    def partitioning(self, use_partitions, update_with_test=False):
        """
        Handle the partitioning system to use and reset the model to perform
        new evaluations

        Parameters
        ----------
        use_partitions: True if train/set partitioning is needed, False
                        otherwise
        update_with_test: True if the model should be updated with the test set,
                          False otherwise
        """
        self.use_partitions = use_partitions
        self.update_with_test = update_with_test
        self.id2word = None
        self.id_corpus = None

    def train_model(self, dataset, hyperparams=None, top_words=10):
        """
        Train the model and return output

        Parameters
        ----------
        dataset : dataset to use to build the model
        hyperparams : hyperparameters to build the model
        top_words : if greater than 0 returns the most significant words for
                    each topic in the output (Default True)
        Returns
        -------
        result : dictionary with up to 3 entries,
                 'topics', 'topic-word-matrix' and
                 'topic-document-matrix'
        """
        if hyperparams is None:
            hyperparams = {}

        if self.use_partitions:
            train_corpus, test_corpus = dataset.get_partitioned_corpus(
                use_validation=False)
        else:
            train_corpus = dataset.get_corpus()

        if self.id2word is None:
            self.id2word = corpora.Dictionary(dataset.get_corpus())

        if self.id_corpus is None:
            self.id_corpus = [self.id2word.doc2bow(document)
                              for document in train_corpus]

        if "num_topics" not in hyperparams:
            hyperparams["num_topics"] = self.hyperparameters["num_topics"]

        # Allow alpha to be a float in case of symmetric alpha
        if "alpha" in hyperparams:
            if isinstance(hyperparams["alpha"], float):
                hyperparams["alpha"] = [
                    hyperparams["alpha"]
                ] * hyperparams["num_topics"]

        hyperparams["corpus"] = self.id_corpus
        hyperparams["id2word"] = self.id2word
        self.hyperparameters.update(hyperparams)

        self.trained_model = ldamodel.LdaModel(**self.hyperparameters)

        result = {}

        result["topic-word-matrix"] = self.trained_model.get_topics()

        if top_words > 0:
            topics_output = []
            for topic in result["topic-word-matrix"]:
                top_k = np.argsort(topic)[-top_words:]
                top_k_words = list(reversed([self.id2word[i] for i in top_k]))
                topics_output.append(top_k_words)
            result["topics"] = topics_output

        result["topic-document-matrix"] = self._get_topic_document_matrix()

        if self.use_partitions:
            new_corpus = [self.id2word.doc2bow(
                document) for document in test_corpus]
            if self.update_with_test:
                self.trained_model.update(new_corpus)
                self.id_corpus.extend(new_corpus)

                result["test-topic-word-matrix"] = (
                    self.trained_model.get_topics())

                if top_words > 0:
                    topics_output = []
                    for topic in result["test-topic-word-matrix"]:
                        top_k = np.argsort(topic)[-top_words:]
                        top_k_words = list(
                            reversed([self.id2word[i] for i in top_k]))
                        topics_output.append(top_k_words)
                    result["test-topics"] = topics_output

                result["test-topic-document-matrix"] = (
                    self._get_topic_document_matrix())

            else:
                test_document_topic_matrix = []
                for document in new_corpus:
                    document_topics_tuples = self.trained_model[document]
                    document_topics = np.zeros(
                        self.hyperparameters["num_topics"])
                    for single_tuple in document_topics_tuples:
                        document_topics[single_tuple[0]] = single_tuple[1]

                    test_document_topic_matrix.append(document_topics)
                result["test-topic-document-matrix"] = np.array(
                    test_document_topic_matrix).transpose()
        return result

    def _get_topics_words(self, topk):
        """
        Return the most significative words for each topic.
        """
        topic_terms = []
        for i in range(self.hyperparameters["num_topics"]):
            topic_words_list = []
            for word_tuple in self.trained_model.get_topic_terms(i, topk):
                topic_words_list.append(self.id2word[word_tuple[0]])
            topic_terms.append(topic_words_list)
        return topic_terms

    def _get_topic_document_matrix(self):
        """
        Return the topic representation of the
        corpus
        """
        doc_topic_tuples = []
        for document in self.id_corpus:
            doc_topic_tuples.append(
                self.trained_model.get_document_topics(document,
                                                       minimum_probability=0))

        topic_document = np.zeros((
            self.hyperparameters["num_topics"],
            len(doc_topic_tuples)))

        for ndoc in range(len(doc_topic_tuples)):
            document = doc_topic_tuples[ndoc]
            for topic_tuple in document:
                topic_document[topic_tuple[0]][ndoc] = topic_tuple[1]
        return topic_document



import importlib
import json
import os
import sys
from importlib import util
from pathlib import Path

import gensim.corpora as corpora
import numpy as np
from skopt.space.space import Real, Categorical, Integer

import octis.configuration.defaults as defaults
from octis.models.model import load_model_output

path = Path(os.path.dirname(os.path.realpath(__file__)))
pathDataset = str(path.parent.parent)
path = str(path.parent)

# Import octis module
spec = importlib.util.spec_from_file_location(
    "octis", str(os.path.join(path, "__init__.py")), submodule_search_locations=[])
module = importlib.util.module_from_spec(spec)
sys.modules[spec.name] = module
spec.loader.exec_module(module)
importlib.invalidate_caches()


def importClass(class_name, module_name, module_path):
    """
    Import a class runtime based on its module and name

    :param class_name: name of the class
    :type class_name: String
    :param module_name: name of the module
    :type module_name: String
    :param module_path: absolute path to the module
    :type module_path: String

    :return: returns the selected class
    :rtype: Object
    """
    spec = importlib.util.spec_from_file_location(
        module_name, module_path, submodule_search_locations=[])
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    importlib.invalidate_caches()
    single_class = getattr(module, class_name)
    return single_class


def importModel(model_name):
    """
    Import a model runtime based on its name

    :param model_name: name of the model
    :type model_name: String

    :return: returns a model.
    :rtype: Model
    """
    module_path = os.path.join(path, "models")
    module_path = os.path.join(module_path, model_name + ".py")
    model = importClass(model_name, model_name, module_path)
    return model


def importMetric(metric_name):
    """
    Import a metric runtime based on its name

    :param metric_name: name of the metric
    :type metric_name: String

    :return: returns a metric
    :rtype: Metric
    """
    module_path = os.path.join(path, "evaluation_metrics")
    module_name = defaults.metric_parameters[metric_name]["module"]
    module_path = os.path.join(module_path, module_name + ".py")
    metric = importClass(metric_name, metric_name, module_path)
    return metric


def importDataset():
    """
    Import the class dataset at runtime

    :return: returns the dataset class
    :rtype: Dataset
    """
    module_path = os.path.join(path, "dataset")
    module_path = os.path.join(module_path, "dataset.py")
    dataset_class = importClass("Dataset", "dataset", module_path)
    return dataset_class


def importOptimizer():
    """
    Import the optimizer at runtime

    :return: returns the oprimizer class
    :rtype: Optimizer
    """
    module_path = os.path.join(path, "optimization")
    module_path = os.path.join(module_path, "optimizer.py")
    optimizer_class = importClass("Optimizer", "optimizer", module_path)
    return optimizer_class


# TO UPDATE
def startExperiment(parameters):
    """
    Starts an experiment with the given parameters

    :param parameters: parameters of the experiment
    :type parameters: Dict
    """

    optimizationPath = str(os.path.join(
        parameters["path"], parameters["experimentId"]))
    json_file = str(os.path.join(optimizationPath,
                                 parameters["experimentId"] + ".json"))
    if os.path.isfile(json_file):
        Optimizer = importOptimizer()
        optimizer = Optimizer()
        optimizer.resume_optimization(json_file)
    else:
        # Import dataset class and initialize an instance with the chosen dataset
        dataset_class = importDataset()
        dataset = dataset_class()
        dataset_path = str(os.path.join(
            pathDataset, "preprocessed_datasets", parameters["dataset"]))
        dataset.load_custom_dataset_from_folder(dataset_path)

        model_class = importModel(parameters["model"]["name"])
        model = model_class()

        model.hyperparameters.update(parameters["model"]["parameters"])
        model.partitioning(parameters["partitioning"])

        search_space = {}

        for key, value in parameters["optimization"]["search_spaces"].items():
            if "low" in value:
                if isinstance(value["low"], float) or isinstance(value["high"], float):
                    search_space[key] = Real(
                        low=value["low"], high=value["high"])
                else:
                    search_space[key] = Integer(
                        low=value["low"], high=value["high"])
            else:
                search_space[key] = Categorical(value)

        metric_parameters = parameters["optimize_metrics"][0]["parameters"]
        for key in metric_parameters:
            if metric_parameters[key] == "use dataset texts":
                metric_parameters[key] = dataset.get_corpus()
            elif metric_parameters[key] == "use selected dataset":
                metric_parameters[key] = dataset
            elif os.path.isdir(str(metric_parameters[key])):
                metricDataset = dataset_class()
                metricDataset.load_custom_dataset_from_folder(
                    metric_parameters[key])
                metric_parameters[key] = metricDataset.get_corpus()

        metric_class = importMetric(parameters["optimize_metrics"][0]["name"])
        metric = metric_class(**metric_parameters)

        metrics_to_track = []
        for single_metric in parameters["track_metrics"]:
            metric_class = importMetric(single_metric["name"])
            single_metric_parameters = single_metric["parameters"]
            for key in single_metric_parameters:
                if single_metric_parameters[key] == "use dataset texts":
                    single_metric_parameters[key] = dataset.get_corpus()
                elif single_metric_parameters[key] == "use selected dataset":
                    single_metric_parameters[key] = dataset
            new_metric = metric_class(**single_metric_parameters)
            metrics_to_track.append(new_metric)

        vocabulary_path = str(os.path.join(
            parameters["path"], parameters["experimentId"], "models"))

        Path(vocabulary_path).mkdir(parents=True, exist_ok=True)

        vocabulary_path = str(os.path.join(vocabulary_path, "vocabulary.json"))

        file = open(vocabulary_path, "w")
        json.dump(dict(corpora.Dictionary(dataset.get_corpus())), file)
        file.close()

        Optimizer = importOptimizer()
        optimizer = Optimizer()
        optimizer.optimize(model, dataset, metric, search_space, metrics_to_track, random_state=True,
                           initial_point_generator="random",
                           surrogate_model=parameters["optimization"]["surrogate_model"],
                           model_runs=parameters["optimization"]["model_runs"],
                           n_random_starts=parameters["optimization"]["n_random_starts"],
                           acq_func=parameters["optimization"]["acquisition_function"],
                           number_of_call=parameters["optimization"]["iterations"],
                           save_models=True, save_name=parameters["experimentId"], save_path=optimizationPath)


def retrieveBoResults(result_path):
    """
    Function to load the results_old of BO

    :param result_path: path where the results_old are saved (json file)
    :type result_path: String

    :return: returns the results of BO
    :rtype: Dict
    """
    if os.path.isfile(result_path):
        # open json file
        with open(result_path, 'rb') as file:
            result = json.load(file)
        f_val = result['f_val']
        # output dictionary
        dict_return = dict()
        dict_return.update({"f_val": f_val,
                            "current_iteration": result["current_call"],
                            "total_iterations": result["number_of_call"]})
        return dict_return
    return False


def retrieveIterationBoResults(path, iteration):
    """
    Function to load the results_old of BO until iteration

    :param path: path where the results_old are saved (json file).
    :type path: String
    :param iteration: considered iteration.
    :type iterations: Int

    :return: returns the BO results until the given iteration
    :rtype: Dict
    """
    if os.path.isfile(path):
        # open json file
        with open(path, 'rb') as file:
            result = json.load(file)
        values = result["f_val"]

        type_of_problem = result['optimization_type']
        hyperparameters = result['x_iters']
        name_hyp = list(hyperparameters.keys())
        hyperparameters_iter = list()
        for name in name_hyp:
            hyperparameters_iter.append(hyperparameters[name][iteration])

        hyperparameters_config = dict(zip(name_hyp, hyperparameters_iter))
        # output dictionary
        dict_return = dict()

        metric_name = result["metric_name"]
        values = result['dict_model_runs'][metric_name]['iteration_' +
                                                        str(iteration)]

        extra_metric_names = result["extra_metric_names"]
        for name in extra_metric_names:
            values = result['dict_model_runs'][name]['iteration_' +
                                                     str(iteration)]
            dict_return.update({name + "_values": values})

        dict_metrics = result['dict_model_runs']
        model_runs = result['model_runs']
        name_metrics = list(dict_metrics.keys())
        if len(result['extra_metric_names']) > 0:
            # metrics names
            dict_return.update({"metric_names": name_metrics[1:]})
        else:
            dict_return.update({"metric_names": 0})

        dict_return.update({"optimized_metric": result["metric_name"]})
        dict_return.update({"optimized_metric_values": values})
        dict_model_attributes = result['model_attributes']
        dict_return.update({"model_attributes": dict_model_attributes})
        dict_return.update({"model_name": result["model_name"]})
        dict_return.update(
            {"hyperparameter_configuration": hyperparameters_config})
        return dict_return
    return False
# Manca retrieve di Iperparametri iterazione e topic-word-matrix e document-topic matrix


def singleInfo(path):
    """
    Compute average, median, best and worst result of the object function evaluations

    :param path: path of the json file of a single experiment
    :type path: String

    :return: average, median, best and worst result of the object function evaluations
    :rtype: Dict
    """
    if os.path.isfile(path):
        with open(path, 'rb') as file:
            result = json.load(file)
        values = result['f_val']
        type_of_problem = result['optimization_type']
        if type_of_problem == 'Maximize':
            best_seen = max(values)
            worse_seen = min(values)
            best_index = np.argmax(values)
        else:
            best_seen = min(values)
            worse_seen = max(values)
            best_index = np.argmin(values)
        median_seen = np.median(values)
        mean_seen = np.mean(values)

        dict_metrics = result['dict_model_runs']
        model_runs = result['model_runs']
        name_metrics = list(dict_metrics.keys())
        dict_results = dict()
        for name in name_metrics:
            dict_results[name] = list()
            metric_results = dict_metrics[name]
            iterations = len(metric_results.keys())
            for i in range(iterations):
                dict_results[name].append(
                    metric_results['iteration_' + str(i)])

        hyperparameters = result['x_iters']
        name_hyp = list(hyperparameters.keys())
        best_hyperparameter_configuration = dict()
        for name in name_hyp:
            best_hyperparameter_configuration.update(
                {name: hyperparameters[name][best_index]})
        # dizionaro di output
        dict_return = dict()
        dict_return.update({"model_runs": dict_results})
        dict_return.update({"f_val": values})
        dict_return.update({"best_seen": best_seen})
        dict_return.update({"worse_seen": worse_seen})
        dict_return.update({"median_seen": median_seen})
        dict_return.update({"mean_seen": mean_seen})
        dict_return.update({"number_of_model_runs": model_runs})
        dict_return.update({"current_iteration": result["current_call"],
                            "total_iterations": result["number_of_call"]})
        dict_return.update({"hyperparameter_configurations": hyperparameters})
        dict_return.update(
            {"hyperparameter_configuration": best_hyperparameter_configuration})
        dict_return.update({"optimized_metric": result["metric_name"]})

        # other hyper-parameter values
        dict_model_attributes = result['model_attributes']
        dict_return.update({"model_attributes": dict_model_attributes})
        dict_return.update({"model_name": result["model_name"]})

        dict_values_extra_metrics = dict()
        dict_stats_extra_metrics = dict()

        if len(result['extra_metric_names']) > 0:
            # metrics names
            dict_return.update({"metric_names": name_metrics[1:]})
            extra_metrics_names = list(result['dict_model_runs'].keys())

            for name in extra_metrics_names[1:]:
                values = []
                dict_values = result['dict_model_runs'][name]
                iterations = list(dict_values.keys())
                for j in iterations:
                    values.append(np.median(dict_values[j]))
                dict_values_extra_metrics.update({name: values})
                val_stats = [np.max(values), np.min(
                    values), np.median(values), np.mean(values)]
                dict_stats_extra_metrics.update({name: val_stats})
            dict_return.update(
                {"extra_metric_vals": dict_values_extra_metrics})
            dict_return.update(
                {"extra_metric_stats": dict_stats_extra_metrics})
        else:
            dict_return.update({"metric_names": 0})
            dict_return.update({"extra_metric_vals": dict()})
            dict_return.update({"extra_metric_stats": dict()})

        return dict_return
    return None


def getModelInfo(experiment_path, iteration, modelRun):
    """
    Retrieve the output of the given model

    :param experiment_path:  path of the experiment folder
    :type experiment_path: String
    :param iteration: number of iteration
    :type iteration: Int
    :param modelRun: number of model run
    :type modelRun: Int

    :return: output of the model and vocabulary
    :rtype: Dict
    """
    outputfile = str(os.path.join(experiment_path,
                                  "models",
                                  str(iteration) + "_" + str(modelRun) + ".npz"))
    vocabularyfile = str(os.path.join(experiment_path,
                                      "models",
                                      "vocabulary.json"))
    if os.path.isfile(outputfile) and os.path.isfile(vocabularyfile):
        output = load_model_output(outputfile, vocabularyfile, 20)
        return output
    return False



from octis.models.model import AbstractModel
import numpy as np
from gensim.models import hdpmodel
import gensim.corpora as corpora
import octis.configuration.citations as citations
import octis.configuration.defaults as defaults


class HDP(AbstractModel):

    id2word = None
    id_corpus = None
    use_partitions = True
    update_with_test = False

    def __init__(self, max_chunks=None, max_time=None, chunksize=256, kappa=1.0, tau=64.0, K=15, T=150, alpha=1,
                 gamma=1, eta=0.01, scale=1.0, var_converge=0.0001):
        """
        Initialize HDP model

        Parameters
        ----------
        max_chunks (int, optional) – Upper bound on how many chunks to process.
        It wraps around corpus beginning in another corpus pass,
        if there are not enough chunks in the corpus.

        max_time (int, optional) – Upper bound on time (in seconds)
        for which model will be trained.

        chunksize (int, optional) – Number of documents in one chuck.

        kappa (float,optional) – Learning parameter which acts as exponential
        decay factor to influence extent of learning from each batch.

        tau (float, optional) – Learning parameter which down-weights
        early iterations of documents.

        K (int, optional) – Second level truncation level

        T (int, optional) – Top level truncation level

        alpha (int, optional) – Second level concentration

        gamma (int, optional) – First level concentration

        eta (float, optional) – The topic Dirichlet

        scale (float, optional) – Weights information from the
        mini-chunk of corpus to calculate rhot.

        var_converge (float, optional) – Lower bound on the right side of
        convergence. Used when updating variational parameters
        for a single document.
        """
        super().__init__()
        self.hyperparameters["max_chunks"] = max_chunks
        self.hyperparameters["max_time"] = max_time
        self.hyperparameters["chunksize"] = chunksize
        self.hyperparameters["kappa"] = kappa
        self.hyperparameters["tau"] = tau
        self.hyperparameters["K"] = K
        self.hyperparameters["T"] = T
        self.hyperparameters["alpha"] = alpha
        self.hyperparameters["gamma"] = gamma
        self.hyperparameters["eta"] = eta
        self.hyperparameters["scale"] = scale
        self.hyperparameters["var_converge"] = var_converge

    def info(self):
        """
        Returns model informations
        """
        return {
            "citation": citations.models_HDP,
            "name": "HDP, Hierarchical Dirichlet Process"
        }

    def hyperparameters_info(self):
        """
        Returns hyperparameters informations
        """
        return defaults.HDP_hyperparameters_info

    def partitioning(self, use_partitions, update_with_test=False):
        """
        Handle the partitioning system to use and reset the model to perform
        new evaluations

        Parameters
        ----------
        use_partitions: True if train/set partitioning is needed, False
                        otherwise
        update_with_test: True if the model should be updated with the test set,
                          False otherwise
        """
        self.use_partitions = use_partitions
        self.update_with_test = update_with_test
        self.id2word = None
        self.id_corpus = None

    def train_model(self, dataset, hyperparameters={}, topics=10):
        """
        Train the model and return output

        Parameters
        ----------
        dataset : dataset to use to build the model
        hyperparameters : hyperparameters to build the model
        topics : if greather than 0 returns the top k most significant
                 words for each topic in the output
                 Default True

        Returns
        -------
        result : dictionary with up to 3 entries,
                 'topics', 'topic-word-matrix' and
                 'topic-document-matrix'
        """
        partition = []
        if self.use_partitions:
            partition = dataset.get_partitioned_corpus()
        else:
            partition = [dataset.get_corpus(), []]

        if self.id2word is None:
            self.id2word = corpora.Dictionary(dataset.get_corpus())

        if self.id_corpus is None:
            self.id_corpus = [self.id2word.doc2bow(
                document) for document in partition[0]]

        hyperparameters["corpus"] = self.id_corpus
        hyperparameters["id2word"] = self.id2word
        self.hyperparameters.update(hyperparameters)

        self.trained_model = hdpmodel.HdpModel(**self.hyperparameters)

        result = dict()
        result["topic-word-matrix"] = self.trained_model.get_topics()

        if topics > 0:
            topics_output = []
            for topic in result["topic-word-matrix"]:
                top_k = np.argsort(topic)[-topics:]
                top_k_words = list(reversed([self.id2word[i] for i in top_k]))
                topics_output.append(top_k_words)
            result["topics"] = topics_output

        result["topic-document-matrix"] = self._get_topic_document_matrix()
        if self.use_partitions:
            new_corpus = [self.id2word.doc2bow(
                document) for document in partition[1]]
            if self.update_with_test:
                self.trained_model.update(new_corpus)
                self.id_corpus.extend(new_corpus)

                result["test-topic-word-matrix"] = self.trained_model.get_topics()

                if topics > 0:
                    topics_output = []
                    for topic in result["test-topic-word-matrix"]:
                        top_k = np.argsort(topic)[-topics:]
                        top_k_words = list(
                            reversed([self.id2word[i] for i in top_k]))
                        topics_output.append(top_k_words)
                    result["test-topics"] = topics_output

                result["test-topic-document-matrix"] = self._get_topic_document_matrix()

            else:
                test_document_topic_matrix = []
                for document in new_corpus:

                    document_topics_tuples = self.trained_model[document]
                    document_topics = np.zeros(
                        len(self.trained_model.get_topics()))
                    for single_tuple in document_topics_tuples:
                        document_topics[single_tuple[0]] = single_tuple[1]

                    test_document_topic_matrix.append(document_topics)

                result["test-topic-document-matrix"] = np.array(
                    test_document_topic_matrix).transpose()

        return result

    def _get_topics_words(self, topics):
        """
        Return the most significative words for each topic.
        """
        topic_terms = []
        for i in range(len(self.trained_model.get_topics())):
            topic_terms.append(self.trained_model.show_topic(
                i,
                topics,
                False,
                True
            ))
        return topic_terms

    def _get_topic_document_matrix(self):
        """
        Return the topic representation of the
        corpus
        """
        doc_topic_tuples = []
        for document in self.id_corpus:
            doc_topic_tuples.append(self.trained_model[document])

        topic_document = np.zeros((
            len(self.trained_model.get_topics()),
            len(doc_topic_tuples)))

        for ndoc in range(len(doc_topic_tuples)):
            document = doc_topic_tuples[ndoc]
            for topic_tuple in document:
                topic_document[topic_tuple[0]][ndoc] = topic_tuple[1]
        return topic_document
